import json
import re

from inspect import isclass
from typing import *

from .typing import GruiModel


def convert_argument_list(type_: type, value: Any) -> List[Any]:
    """
    Convert any value to a list.

    :param type_:  The type of the element contained into the list.
    :param value:  The value to convert to the right type.

    :return:  The list converted to the right type
    """

    # Initialize the result
    result = []
    # First wz check if we already have a list.
    if isinstance(value, (list, tuple)):
        for tmp in value:
            result.append(convert_argument(type_, tmp, False))
    # Maybe this is a list of value separated by comma
    elif isinstance(value, str):
        for tmp in value.split(","):
            result.append(convert_argument(type_, tmp, False))
    else:
        result.append(convert_argument(type_, value, False))
    return result


def convert_argument(type_: type, value: Any, to_list: bool) -> Any:
    if to_list:
        return convert_argument_list(type_, value)
    else:
        if type_ is int:
            return int(value)
        elif type_ is float:
            return float(value)
        elif type_ is bool:
            return bool(value)
        elif get_origin(type_) == list:
            return convert_argument_list(get_args(type_)[0], value)
        elif isclass(type_) and issubclass(type_, GruiModel):
            return type_.from_json(value)
    return value


def dict2list(dict_: dict) -> List[dict]:
    """:return A list of dict where every dict contains one value"""
    result = []
    for key in dict_:
        for i, value in enumerate(dict_[key]):
            try:
                result[i][key] = value
            except IndexError:
                result.append({key: value})
    return result


class GruiJsonEncoder(json.JSONEncoder):

    __DICT_VALUES_TYPE = type({}.values())
    __DICT_KEYS_TYPE = type({}.keys())

    def default(self, obj):
        from .typing import GruiJsonSerializable

        if isinstance(obj, GruiJsonEncoder.__DICT_VALUES_TYPE) or isinstance(obj, GruiJsonEncoder.__DICT_KEYS_TYPE):
            return tuple(obj)
        elif isinstance(obj, GruiJsonSerializable):
            return obj.to_json()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


_REGEX_HAVE_DIGIT = re.compile("\\d")
_REGEX_HAVE_UPPER = re.compile("[a-z]")
_REGEX_HAVE_LOWER = re.compile("[A-Z]")
_REGEX_HAVE_SYMBOL = re.compile("[!#:;$%&'()*+,\\]\\-/[^_`}{|~\"]")


def is_strong_password(password: str):
    # Check the length
    if len(password) < 8:
        return False
    # Search for at least one digit
    if _REGEX_HAVE_DIGIT.search(password) is None:
        return False
    # Search for at least one upper case
    if _REGEX_HAVE_UPPER.search(password) is None:
        return False
    # Search for at least one lower case
    if _REGEX_HAVE_LOWER.search(password) is None:
        return False
    # Search for at least one symbol
    if _REGEX_HAVE_SYMBOL.search(password) is None:
        return False
    return True



import typing
import re
import inspect
import docstring_parser

from typing import *
from abc import ABC, abstractmethod
from case_convert import snake_case, camel_case


##########
# Errors #
##########
class UpdateIdError(RuntimeError):
    def __init__(self, message="Trying to update the id", **kwargs):
        super().__init__(message, kwargs)


class IncorrectDataWhileEncodingError(RuntimeError):
    def __init__(self, message="Trying to encode with incorrect data", wrong_model=None, **kwargs):
        tmp = {}
        if wrong_model:
            for key in wrong_model._GruiModel__initial_attr:
                tmp[key] = wrong_model.__getattribute__(key)
        super().__init__(message, tmp, **kwargs)


class DuplicateTypeError(RuntimeError):
    pass


##############
# Interfaces #
##############
class GruiJsonSerializable(ABC):
    """
    This class is an abstract one. It works like a interface. It simply consist to declare the function to_json.
    Like that an simple check isinstance() is enough to safely serialize to the Json format.
    """
    @abstractmethod
    def to_json(self) -> Dict[str, Any]:
        pass


################
# Meta classes #
################
class _GruiModelMeta(type(GruiJsonSerializable)):
    """
    Meta class to implement the grui_attributes property to any class inherited from 'GruiModel'
    """
    __grui_attributes = {}

    @property
    def grui_attributes(cls) -> Dict[str, "GruiTypeModelAttribute"]:
        from .decorator import GruiFunction

        if cls.__grui_attributes.get(cls) is None:
            cls.__grui_attributes[cls] = {}
            signature = inspect.signature(cls.__init__)
            for name, type_ in get_type_hints(cls).items():
                attr = GruiTypeModelAttribute()
                attr.type = GruiType.get_instance(type_)
                attr.read_only = name == "id"
                attr.description = ""
                try:
                    tmp = signature.parameters[name].default
                    attr.default_value = tmp if tmp != inspect.Parameter.empty else None
                except KeyError:
                    attr.default_value = None
                cls.__grui_attributes[cls][name] = attr

            if cls != GruiFunction and cls.__doc__ is not None:
                doc_parsed = docstring_parser.parse(cls.__doc__)
                for attribute in doc_parsed.params:
                    if attribute.arg_name in cls.__grui_attributes[cls].keys() and attribute.description is not None:
                        cls.__grui_attributes[cls][attribute.arg_name].description = attribute.description

        return cls.__grui_attributes[cls]


#############
# Variables #
#############

NoneType = type(None)


###########
# Classes #
###########

class GruiModel(GruiJsonSerializable, metaclass=_GruiModelMeta):
    """
    Base class of the of any object model saved and use by the application.
    This class allow developer to focus mainly on how the data it saved and related to eachother.
    The class require to have a id. It type can be change but I always recommend to set it optional.
    """

    ##############
    # Attributes #
    ##############
    id: Optional[int]

    ###############
    # Constructor #
    ###############
    def __init__(self, **kwargs):
        self.__id = None
        self.__json_result = None
        self.__initial_attr = kwargs.keys()
        for attr_name, attr_value in kwargs.items():
            self.__setattr__(attr_name, attr_value) 

    ##########
    # Getter #
    ##########
    @property
    def id(self):
        return self.__id

    ##########
    # Setter #
    ##########
    @id.setter
    def id(self, id_: int):
        if self.__id is None:
            self.__id = id_
        else:
            raise UpdateIdError()

    def __setattr__(self, key, value):
        self.__dict__["_GruiModel__json_result"] = None
        super().__setattr__(key, value)

    ###########
    # Methods #
    ###########
    def to_json(self):
        if not self.validate():
            raise IncorrectDataWhileEncodingError(self.__class__.__name__)
        if self.__json_result is None:
            json_result = {}
            for attr_name, attr in self.__class__.grui_attributes.items():
                json_result[camel_case(attr_name)] = attr.type.jsonify_value(self.__getattribute__(attr_name))
            self.__json_result = json_result
        return self.__json_result

    def validate(self) -> bool:
        for attr_name, attr in self.__class__.grui_attributes.items():
            try:
                attr_value = self.__getattribute__(attr_name)
            except AttributeError:
                attr_value = None
                self.__setattr__(attr_name, attr_value)
            valid, attr_validated = attr.type.validate_value(attr_value)
            if not valid:
                return False
            elif attr_validated != attr_value:
                if attr_name == "id":
                    self.__dict__["_GruiModel__id"] = attr_validated
                else:
                    self.__setattr__(attr_name, attr_validated)
        return True

    @classmethod
    def from_json(cls, values: Any) -> "GruiModel":
        parameters = {}
        if isinstance(values, dict):
            for key in values:
                parameters[snake_case(key)] = values[key]
            return cls(**parameters)
        else:
            return cls(values)

    def update(self, other: "GruiModel"):
        updated = self.__class__(**self.__dict__)
        for key in other.__initial_attr:
            updated.__setattr__(key, other.__getattribute__(key))
        return updated


class GruiSubModel(GruiJsonSerializable, metaclass=_GruiModelMeta):

    ###############
    # Constructor #
    ###############
    def __init__(self, **kwargs):
        self.__json_result = None
        for attr_name, attr_value in kwargs.items():
            self.__setattr__(attr_name, attr_value)

    ##########
    # Setter #
    ##########
    def __setattr__(self, key, value):
        self.__dict__["_GruiSubModel__json_result"] = None
        super().__setattr__(key, value)

    ##########
    # Method #
    ##########
    def to_json(self):
        if self.__json_result is None:
            json_result = {}
            for attr_name, attr in self.__class__.grui_attributes.items():
                json_result[camel_case(attr_name)] = attr.type.jsonify_value(self.__getattribute__(attr_name))
            self.__json_result = json_result
        return self.__json_result


class Text(str):
    pass


class GruiType(GruiModel):
    """
    It the different data types the application can handle.
    This class shouldn't be instanced directly. Instead use the 'get_instance' with the int of data you want handle.
    """

    ###################
    # Class variables #
    ###################
    __instance = {}

    ##############
    # Attributes #
    ##############
    id: int
    name: str
    primitive: bool
    model: bool
    container: bool
    multiple: bool
    nullable: bool

    #########################
    # Constructor / Builder #
    #########################
    def __new__(cls, type_):
        if cls == GruiType:
            raise RuntimeError("Trying to initiate wrong instance")
        if GruiType.__instance.get(type_) is None:
            GruiType.__instance[type_] = object.__new__(cls)
        elif GruiType.__instance[type_].__class__ != cls:
            raise RuntimeError("Trying to override grui type")
        return GruiType.__instance[type_]

    def __init__(self, original_type):
        super().__init__(id=id(self))
        self._original_type = original_type
        self.primitive = self.model = self.container = self.multiple = self.nullable = False

    ###########
    # Getters #
    ###########
    @staticmethod
    def get_instance(type_):
        """
        Return the right GruiType for the type pass in parameter.
        This function allow us to also have only one instance by type.
        Indeed it is not require to have different instance for the same type.
        """
        if GruiType.__instance.get(type_) is None:
            if type_ == NoneType:
                GruiType.__instance[type_] = GruiTypeNull(type_)
            elif type_ == bool:
                GruiType.__instance[type_] = GruiTypeBool(type_)
            elif type_ == int:
                GruiType.__instance[type_] = GruiTypeInt(type_)
            elif type_ == float:
                GruiType.__instance[type_] = GruiTypeFloat(type_)
            elif inspect.isclass(type_) and issubclass(type_, Text):
                GruiType.__instance[type_] = GruiTypeText(type_)
            elif type_ == str:
                GruiType.__instance[type_] = GruiTypeStr(type_)
            elif get_origin(type_) == list:
                GruiType.__instance[type_] = GruiTypeArray(type_)
            elif get_origin(type_) == dict:
                GruiType.__instance[type_] = GruiTypeMap(type_)
            elif get_origin(type_) == tuple:
                GruiType.__instance[type_] = GruiTypeFixedArray(type_)
            elif get_origin(type_) == Union:
                if len(get_args(type_)) == 2 and get_args(type_)[1] == NoneType:
                    tmp = GruiType.get_instance(get_args(type_)[0])
                    GruiType.__instance[type_] = tmp.__class__(type_)
                    GruiType.__instance[type_].nullable = True
                else:
                    GruiType.__instance[type_] = GruiTypeMultiple(type_)
            elif inspect.isclass(type_) and issubclass(type_, GruiModel):
                GruiType.__instance[type_] = GruiTypeModel(type_)
            elif inspect.isclass(type_) and issubclass(type_, GruiSubModel):
                GruiType.__instance[type_] = GruiTypeSubModel(type_)
            else:
                GruiType.__instance[type_] = GruiTypeUnknown(type_)

        return GruiType.__instance[type_]

    ###########
    # Methods #
    ###########
    def validate(self):
        return True

    @abstractmethod
    def validate_value(self, value):
        pass

    def jsonify_value(self, value):
        return value

    def register(self, register: Dict[int, "GruiType"]):
        if register.get(self.id) is not None and register[self.id] != self:
            raise DuplicateTypeError
        register[self.id] = self

    def __eq__(self, other):
        if self.name == "__unknown__":
            return self.name == other.name
        else:
            return super(GruiType, self).__eq__(other)


class GruiTypeUnknown(GruiType):

    def __init__(self, unknown_type):
        super().__init__(unknown_type)
        self.name = "__unknown__"

    def validate_value(self, value) -> Tuple[bool, Any]:
        return True, value


class GruiTypeNull(GruiType):

    def __init__(self, original_type):
        super().__init__(original_type)
        self.name = "__null__"
        self.nullable = self.primitive = True

    def validate_value(self, value) -> Tuple[bool, None]:
        if value is None:
            return True, None
        return False, value


class GruiTypeInt(GruiType):

    __REGEX_NUMBER_10 = re.compile("^\\d+$")
    __REGEX_NUMBER_16 = re.compile("^0x[0-9a-f]+$")
    __REGEX_NUMBER_8 = re.compile("^0o[0-7]+$")
    __REGEX_NUMBER_2 = re.compile("^0b[0-1]+$")

    def __init__(self, original_type):
        super().__init__(original_type)
        self.primitive = True
        self._original_type = original_type
        self.name = "__int__"

    def validate_value(self, value) -> Tuple[bool, Optional[int]]:
        if self.nullable and value is None:
            return True, value
        if type(value) is int:
            return True, value
        elif type(value) is str:
            if GruiTypeInt.__REGEX_NUMBER_10.match(value):
                return True, int(value)
            elif GruiTypeInt.__REGEX_NUMBER_16.match(value.lower()):
                return True, int(value, 16)
            elif GruiTypeInt.__REGEX_NUMBER_8.match(value):
                return True, int(value, 8)
            elif GruiTypeInt.__REGEX_NUMBER_2.match(value):
                return True, int(value, 2)
        return False, value


class GruiTypeFloat(GruiType):

    __REGEX_FLOAT = re.compile("^\\d*\\.\\d+$")

    def __init__(self, original_type):
        super().__init__(original_type)
        self.primitive = True
        self.name = "__float__"

    def validate_value(self, value) -> Tuple[bool, Optional[float]]:
        if self.nullable and value is None:
            return True, value
        if type(value) is float:
            return True, value
        elif type(value) is int:
            return True, float(value)
        elif type(value) is str and GruiTypeFloat.__REGEX_FLOAT.match(value):
            return True, float(value)
        return False, value


class GruiTypeBool(GruiType):

    def __init__(self, original_type):
        super().__init__(original_type)
        self.primitive = True
        self.name = "__bool__"

    def validate_value(self, value) -> Tuple[bool, Optional[bool]]:
        if self.nullable and value is None:
            return True, None
        elif type(value) is bool:
            return True, value
        return False, value


class GruiTypeStr(GruiType):

    def __init__(self, original_type):
        super().__init__(original_type)
        self.primitive = True
        self.name = "__str__"

    def validate_value(self, value) -> Tuple[bool, Optional[str]]:
        if self.nullable and value is None:
            return True, None
        if type(value) is str:
            return value != "", value
        return False, value


class GruiTypeText(GruiTypeStr):

    def __init__(self, original_type):
        super().__init__(original_type)
        self.name = "__text__"


class GruiTypeModelAttribute(GruiSubModel):

    read_only: bool
    type: GruiType
    default_value: Any
    description: str


class GruiTypeModel(GruiType):

    children: Dict[str, GruiTypeModelAttribute]

    def __init__(self, original_type):
        super().__init__(original_type)
        if get_origin(original_type) == Union:
            self.name = snake_case(get_args(original_type)[0].__name__)
        else:
            self.name = snake_case(original_type.__name__)
        self.model = True
        self.__children = None

    @property
    def children(self) -> Dict[str, GruiTypeModelAttribute]:
        return get_args(self._original_type)[0].grui_attributes \
            if self.nullable else self._original_type.grui_attributes

    def jsonify_value(self, value):
        return None if value is None else value.id

    def register(self, register: Dict[int, GruiType]):
        super().register(register)
        for attr in self.children.values():
            attr.type.register(register)

    def validate_value(self, value) -> Tuple[bool, Optional[GruiModel]]:
        if self.nullable and value is None:
            return True, None
        if type(value) is self._original_type:
            return True, value
        return False, value


class GruiTypeSubModel(GruiType):

    children: Dict[str, GruiTypeModelAttribute]

    def __init__(self, original_type):
        super().__init__(original_type)
        if get_origin(original_type) == Union:
            self.name = snake_case(get_args(original_type)[0].__name__)
        else:
            self.name = snake_case(original_type.__name__)
        self.model = True

    @property
    def children(self) -> Dict[str, GruiTypeModelAttribute]:
        return self._original_type.grui_attributes

    def jsonify_value(self, value):
        return None if value is None else value.to_json()

    def validate_value(self, value) -> Tuple[bool, Optional[GruiModel]]:
        if self.nullable and value is None:
            return True, None
        if type(value) is self._original_type:
            return True, value
        return False, value


class GruiTypeArray(GruiType):

    child: GruiType

    def __init__(self, original_type):
        super().__init__(original_type)
        self.child = GruiType.get_instance(get_args(original_type)[0])
        self.name = "__array__(" + self.child.name + ")"
        self.container = True
        self.primitive = self.child.primitive
        self.model = self.child.model
        self.jsonify = 44

    def jsonify_value(self, value):
        result = []
        for tmp in value:
            result.append(self.child.jsonify_value(tmp))
        return result

    def register(self, register: Dict[int, "GruiType"]):
        super().register(register)
        self.child.register(register)

    def validate_value(self, value) -> Tuple[bool, Optional[list]]:
        if self.nullable and value is None:
            return True, None
        if type(value) is list:
            for i, tmp in enumerate(value):
                child_valid, child_value = self.child.validate_value(tmp)
                if not child_valid:
                    return False, value
                value[i] = child_value
            return True, value
        return False, value


class GruiTypeMap(GruiType):

    key_type: GruiType
    value_type: GruiType

    def __init__(self, original_type):
        super().__init__(original_type)
        if get_origin(original_type) == Union and \
                len(get_args(original_type)) == 2 and \
                get_args(original_type)[1] == NoneType:
            self.nullable = True
            self.key_type = GruiType.get_instance(get_args(get_args(original_type)[0])[0])
            self.value_type = GruiType.get_instance(get_args(get_args(original_type)[0])[1])
        else:
            self.key_type = GruiType.get_instance(get_args(original_type)[0])
            self.value_type = GruiType.get_instance(get_args(original_type)[1])
        self.name = "__map__(" + self.key_type.name + ", " + self.value_type.name + ")"
        self.container = True

    def jsonify_value(self, value):
        if value is None:
            return None
        result = {}
        for key, map_value in value.items():
            result[self.key_type.jsonify_value(key)] = self.value_type.jsonify_value(map_value)
        return result

    def register(self, register: Dict[int, "GruiType"]):
        super().register(register)
        self.key_type.register(register)
        self.value_type.register(register)

    def validate_value(self, dict_) -> Tuple[bool, Optional[dict]]:
        if self.nullable and dict_ is None:
            return True, None
        if type(dict_) is dict:
            for key, value in dict_.items():
                key_valid, key = self.key_type.validate_value(key)
                value_valid, value = self.value_type.validate_value(value)
                if not key_valid or not value_valid:
                    return False, value
                value[key] = value
            return True, dict_
        return False, dict_


class GruiTypeFixedArray(GruiType):

    children: Tuple[GruiType, ...]

    def __init__(self, original_type):
        super().__init__(original_type)
        children = []
        name = "__fixed_array__("
        for arg in get_args(original_type):
            tmp = GruiType.get_instance(arg)
            children.append(tmp)
            name += tmp.name + ", "
        self.children = tuple(children)
        self.name = name[:-2] + ")"

    def jsonify_value(self, value):
        result = []
        for i, tmp in enumerate(value):
            result.append(self.children[i].jsonify_value(tmp))
        return result

    def register(self, register: Dict[int, "GruiType"]):
        super().register(register)
        for child in self.children:
            child.register(register)

    def validate_value(self, value) -> Tuple[bool, Optional[tuple]]:
        if self.nullable and value is None:
            return True, None
        if type(value) is tuple and len(value) == len(self.children):
            for i, tmp in enumerate(value):
                child_valid, child_value = self.children[i].validate(tmp)
                if not child_valid:
                    return False, value
                value[i] = child_value
            return True, value
        return False, value


class GruiTypeMultiple(GruiType):

    choices: Tuple[GruiType, ...]

    def __init__(self, multiple_type):
        super().__init__(multiple_type)
        self.primitive = True
        self.multiple = True
        choices = []
        name = "__multiple__("
        for arg in get_args(multiple_type):
            if arg == NoneType:
                self.nullable = True
            else:
                child = GruiType.get_instance(arg)
                if not child.primitive:
                    self.primitive = False
                if child.model:
                    self.model = True
                choices.append(child)
                name += child.name + ", "
        self.name = name[:-2] + ")"
        self.choices = tuple(choices)

    def jsonify_value(self, value):
        for choice in self.choices:
            if isinstance(value, choice._original_type):
                return choice.jsonify_value(value)
        return value

    def register(self, register: Dict[int, "GruiType"]):
        super().register(register)
        for choice in self.choices:
            choice.register(register)

    def validate_value(self, value) -> Tuple[bool, Optional[Any]]:
        if self.nullable and value is None:
            return True, None
        for choice in self.choices:
            valid, tmp = choice.validate_value(value)
            if valid:
                return True, tmp
        return False, value



from .service import GruiService
from .app import GruiApp, LogLevel
from .decorator import Get, Put, Post, Delete, NotFoundIfNone, NotFoundIfEmpty, MultipleCallAtOnce
from .typing import GruiModel, GruiType, Text, IncorrectDataWhileEncodingError



import docstring_parser

from typing import List
from case_convert import kebab_case

from .typing import GruiType


# This class allow us to generate the url path from the name of the class
# If the developer want to specify a specific name he can use the _url_prefix class variable
class _GruiServiceMeta(type):

    __generated_doc = {}

    @property
    def url_prefix(cls):
        try:
            name = cls.__dict__["_url_prefix"]
        except KeyError:
            name = kebab_case(cls.__name__)
            if name.endswith('-service'):
                name = name[:-8]
        return name

    def generate_doc(cls, app):
        if _GruiServiceMeta.__generated_doc.get(cls) is None:
            docstring_parsed = docstring_parser.parse(cls.__doc__)
            _GruiServiceMeta.__generated_doc[cls] = {
                'name': cls.url_prefix,
                'methods': [],
                'presentation': docstring_parsed.short_description,
                'description': docstring_parsed.long_description,
            }

            for func_id in app.__registered_function:
                func_doc = app.__registered_function[func_id]
                if func_doc.path.startswith(cls.url_prefix, 1):
                    _GruiServiceMeta.__generated_doc[cls]["methods"].append(func_id)
        return _GruiServiceMeta.__generated_doc.get(cls)


class GruiService(metaclass=_GruiServiceMeta):
    pass


class GruiDocService(GruiService):
    from .decorator import _Grui, NotFoundIfEmpty, NotFoundIfNone, MultipleCallAtOnce
    from .app import GruiApp

    _url_prefix = "grui"

    def __init__(self, app: GruiApp):
        self.__app = app

    @_Grui("title")
    def get_title(self):
        return self.__app.title

    @_Grui("function/all")
    @NotFoundIfEmpty
    def get_all_function_doc(self):
        return self.__app.registered_function.values()

    @_Grui("function/<id_>")
    @NotFoundIfNone
    @MultipleCallAtOnce
    def get_function_doc(self, id_: int):
        return self.__app.registered_function.get(id_, None)

    @_Grui("service/all")
    @NotFoundIfEmpty
    def get_all_service_doc(self):
        return self.__app.service_doc.values()

    @_Grui("service/<name>")
    @NotFoundIfNone
    @MultipleCallAtOnce
    def get_service_doc(self, name: str):
        return self.__app.service_doc.get(name, None)

    @_Grui("type/all")
    @NotFoundIfEmpty
    def get_all_type_doc(self) -> List[GruiType]:
        return self.__app.registered_type.values()

    @_Grui("type/<name>")
    @NotFoundIfNone
    @MultipleCallAtOnce
    def get_type_doc(self, name: str) -> GruiType:
        return self.__app.registered_type.get(name, None)



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy import Column, Boolean, Integer, Float, String, JSON
from sqlalchemy.ext.declarative import declarative_base

from typing import *

from .typing import GruiModel, GruiTypeModelAttribute


class DataBase:

    def __init__(self, engine):
        self.__engine = engine
        self.__db_session = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=engine))
        self.__Base = declarative_base()
        self.__mapped_classes = {}

    def bind_class(self, model_class: Type[GruiModel]):

        def init(s, model_instance: GruiModel):
            for attr in model_instance.__class__.grui_attributes:
                s.__setattr__(attr, model_instance.__getattribute__(attr))

        def repr_(s):
            return '<%s %r>' % (s.__class__.__name__, s.id)

        name = "%s.%s" % (model_class.__module__, model_class.__name__)
        attributes = {
            '__tablename__': name,
            '__init__': init,
            '__repr__': repr_
        }
        for attr_name, grui_attr in model_class.grui_attributes.items():
            grui_type = grui_attr.type
            attributes[attr_name] = {
                'nullable': grui_type.nullable,
                'default': grui_attr.default_value
            }
            if attr_name == "id":
                attributes[attr_name]['primary_key'] = True
                attributes[attr_name]['nullable'] = False
                # The following line is done by default by sqlalchemy
                # attributes[attr_name]['autoincrement'] = True
            if grui_type.name == "__bool__":
                return Column(Boolean, **argument)
            elif grui_type.name == "__int__":
                return Column(Integer, **argument)
            elif grui_type.name == "__float__":
                return Column(Float, **argument)
            elif grui_type.name == "__str__":
                return Column(String(150), **argument)
            else:
                return Column(JSON(500), **argument)

        self.__mapped_classes[model_class] = type(name + "Mapped", (self.__Base,), attributes)

    def bind(self):
        self.__Base.query = self.__db_session.query_property()
        self.__Base.metadata.create_all(bind=self.__engine)

    def select_all(self, model_class: Type[GruiModel]):
        return self.__mapped_classes[model_class].query.all()

    def add(self, model_instance: GruiModel):
        mapped_class = self.__mapped_classes[model_instance.__class__]

        mapped_instance = mapped_class(model_instance)
        self.__db_session.add(mapped_instance)

    def commit(self):
        self.__db_session.commit()

    @staticmethod
    def sqlite_engine(file_path):
        return create_engine('sqlite:///'+file_path, convert_unicode=True)



import inspect
import json
import os
import docstring_parser

from functools import wraps
from typing import *
from types import ModuleType
from enum import Enum
from dotenv import dotenv_values
from case_convert import upper_case
from flask import Flask, request, Response, send_file, send_from_directory
from flask_cors import CORS

from .decorator import GruiFunction
from .service import GruiService
from .typing import GruiModel, IncorrectDataWhileEncodingError
from .utils import convert_argument, GruiJsonEncoder, is_strong_password


##########
# Errors #
##########

class InvalidConfigError(RuntimeError):
    pass


#############
# Functions #
#############


def _find_argument(func: GruiFunction, path_parameters: Dict[str, str], body_parameters: Any) -> Dict[str, Any]:
    parameters_found = {}
    signature = inspect.signature(func.callable)

    annotations = dict(signature.parameters.items())
    if func.is_method:
        del annotations["self"]

    if len(annotations) == 0:
        yield {}

    nb_parameter = None

    for param_name in annotations:
        if param_name in path_parameters:
            tmp = convert_argument(annotations[param_name].annotation, path_parameters[param_name], func.multiple_action)
            if nb_parameter is not None and len(tmp) != nb_parameter:
                raise ValueError
            else:
                nb_parameter = len(tmp)
            parameters_found[param_name] = tmp

    if len(parameters_found.keys()) != len(annotations) and body_parameters is not None:
        missing_parameters = [x for x in annotations.keys() if x not in parameters_found.keys()]
        if len(missing_parameters) == 1:
            # Auto convert the body parameter to the missing param
            tmp = convert_argument(annotations[missing_parameters[0]].annotation, body_parameters, func.multiple_action)
            if nb_parameter is not None and len(tmp) != nb_parameter:
                raise ValueError
            else:
                nb_parameter = len(tmp)
            parameters_found[missing_parameters[0]] = tmp

    result = {}
    if func.multiple_action:
        for index in range(nb_parameter):
            for key in parameters_found:
                result[key] = parameters_found[key][index]
            yield result
    else:
        yield parameters_found


def _build4flask(func: GruiFunction):
    @wraps(func)
    def __inner(*args, **kwargs):
        try:
            code = func.default_return_code
            result = []
            for method_arguments in _find_argument(func, kwargs, request.json):
                for previous_action in func.previous_actions:
                    previous_action(code)
                res = func(*args, **method_arguments)
                for next_action in func.next_actions:
                    (res, code) = next_action(res, code)
                if res is None:
                    return Response(json.dumps(res, cls=GruiJsonEncoder), status=code, content_type="application/json")
                if func.multiple_action:
                    result.append(res)
                else:
                    return Response(json.dumps(res, cls=GruiJsonEncoder), status=code, content_type="application/json")
            return Response(json.dumps(result, cls=GruiJsonEncoder), status=code, content_type="application/json")
        except IncorrectDataWhileEncodingError as e:
            return Response(json.dumps(str(e), cls=GruiJsonEncoder), status=400, content_type="application/json")
    return __inner


###########
# Classes #
###########

class LogLevel(Enum):
    """
    Define the log level. Useful to sort the message printed.
    """
    NONE = 0
    WARN = 1
    INFO = 2
    ALL = 3


class GruiApp:

    def __init__(self, title: str = None, service_package: ModuleType = None):
        from .service import GruiDocService
        self.__flask_app = None
        self.__config = {}
        self.load_config()
        self.title = title if title is not None else os.path.split(os.getcwd())[-1]
        self.registered_service = {}
        self.registered_function = {}
        self.registered_type = {}
        self.service_doc = {}
        self.register_service(GruiDocService)
        if service_package is not None:
            self.register_package(service_package)

    @property
    def built(self) -> bool:
        return self.__flask_app is not None

    @property
    def config(self) -> Dict[str, Any]:
        if self.built:
            return self.__flask_app.config
        else:
            return self.__config

    def register_package(self, package: ModuleType):
        """
        This function will look at every member of the module.
        If some are subclass of 'GruiService'. They'll be be added to application.
        """
        for name, member in inspect.getmembers(package, lambda c: inspect.isclass(c) and issubclass(c, GruiService)):
            self.register_service(member)

    def register_service(self, service_class: Type[GruiService]):
        """
        Register every decorated function from the service class.
        """
        if self.registered_service.get(service_class) is not None:
            return

        service_init_arguments = []
        signature = inspect.signature(service_class.__init__)
        for name, parameter in signature.parameters.items():
            if isinstance(self, parameter.annotation):
                service_init_arguments.append(self)
            elif issubclass(parameter.annotation, GruiService):
                self.register_service(parameter.annotation)
                service_init_arguments.append(self.registered_service[parameter.annotation])
        service_instance = service_class(*service_init_arguments)
        self.registered_service[service_class] = service_instance

    def register_function(self, function: GruiFunction):
        """
        Register the function to the application.
        """
        self.registered_function[function.id] = function
        if function.result is not None:
            function.result.type.register(self.registered_type)
        for argument in function.args.values():
            argument.type.register(self.registered_type)
        for service_class in self.registered_service.keys():
            if self.service_doc.get(service_class.url_prefix) is not None and \
                    function.path.startswith(self.config["URL_PREFIX"] + service_class.url_prefix):
                self.service_doc[service_class.url_prefix].methods.append(function)

    def build_wsgi(self, import_name: str, logger_override=None, **kwargs):
        from .service import GruiDocService

        if not self.built:
            self.__config.setdefault('ENVIRONMENT', "development")
            self.__config.setdefault('INDEX_FILE_PATH', "./index.html")
            self.__config.setdefault('FAVICON_FILE_PATH', "./favicon.ico")
            self.__config.setdefault('STATIC_FOLDER_PATH', "static")
            self.__config.setdefault('URL_PREFIX', "/api")

            self.__flask_app = Flask(import_name, **kwargs)
            self.__flask_app.config.update(self.__config)
            self.__flask_app.config['ENV'] = self.__config.get('ENVIRONMENT')
            self.__flask_app.config['FLASK_ENV'] = self.__config.get('ENVIRONMENT')

            if os.path.isfile(self.config['INDEX_FILE_PATH']):
                self.__flask_app.route('/')(self.serve_index_file)

            if os.path.isfile(self.config['FAVICON_FILE_PATH']):
                self.__flask_app.route('/favicon')(self.serve_favicon_file)

            if os.path.isdir(self.config['STATIC_FOLDER_PATH']):
                self.__flask_app.route('/static/<path:path>')(self.serve_static_file)

            if self.config['URL_PREFIX'][-1] != '/':
                self.config['URL_PREFIX'] += "/"

            if not self.validate_config():
                raise InvalidConfigError(self.config)

            if logger_override is not None:
                self.__flask_app.logger.handlers = logger_override.handlers
                self.__flask_app.logger.setLevel(logger_override.level)

            # Register every decorated function into the flask application
            for service_class, service_instance in self.registered_service.items():
                if service_class != GruiDocService:
                    self.service_doc[service_class.url_prefix] = _GruiServiceDoc(service_class)
                for _, method in inspect.getmembers(service_instance, lambda m: isinstance(m, GruiFunction)):
                    self.register_function(method)

            CORS(self.__flask_app, resources={r"/*": {"origins": "*", "methods": ["GET", "POST", "PUT", "DELETE", "GRUI"]}})
            self.__flask_app.url_map.strict_slashes = False
            self.__flask_app.route('/shutdown', methods=['GET'])(GruiApp.shutdown_server)
            self.log(LogLevel.INFO, "Starting application in", self.config['ENVIRONMENT'])

            for function in self.registered_function.values():
                self.log(LogLevel.INFO, "register", "%-7s" % function.method, function.path)
                self.__flask_app.route(function.path, methods=[function.method])(_build4flask(function))

        return self.__flask_app

    def log(self, level: LogLevel, *args):
        log_level = level.name.lower()
        if log_level in dir(self.__flask_app.logger):
            getattr(self.__flask_app.logger, log_level)(" ".join(args))

    def run(self, **kwargs):
        if not self.built:
            self.build_wsgi('__main__')
        self.__flask_app.run(**kwargs)

    def load_config(self):
        for name, value in dotenv_values("./.env").items():
            self.__config[name] = value

    def validate_config(self) -> bool:
        if self.config['ENVIRONMENT'] == "production":
            secret_key = self.config.get("SECRET_KEY", None)
            return type(secret_key) == str and len(secret_key) >= 36 and is_strong_password(secret_key)
        return True

    @staticmethod
    def shutdown_server():
        func = request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()

    def serve_index_file(self):
        return send_file(self.config['INDEX_FILE_PATH'])

    def serve_favicon_file(self):
        return send_file(self.config['FAVICON_FILE_PATH'])

    def serve_static_file(self, path: str):
        return send_from_directory(self.config["STATIC_FOLDER_PATH"], path)


class _GruiServiceDoc(GruiModel):

    id: str
    name: str
    methods: List[GruiFunction]
    presentation: str
    description: str

    def __init__(self, service_class):
        doc_parsed = docstring_parser.parse(service_class.__doc__)
        super().__init__(id=service_class.url_prefix,
                         name=upper_case(service_class.__name__),
                         methods=[],
                         presentation=doc_parsed.short_description or "",
                         description=doc_parsed.long_description or "")

    def validate(self) -> bool:
        return True



import docstring_parser
import re
import inspect

from abc import ABC, abstractmethod
from case_convert import snake_case, upper_case
from typing import *

from .typing import GruiModel, GruiSubModel, GruiType


##########
# Errors #
##########
class InitialisationFunctionError(RuntimeError):
    pass


####################
# Module variables #
####################
MAX_ARGUMENT_PATTERN = re.compile("\\(max: (\\d+.?\\d*)\\)", re.IGNORECASE)
MIN_ARGUMENT_PATTERN = re.compile("\\(min: (\\d+.?\\d*)\\)", re.IGNORECASE)


###########
# Classes #
###########
class GruiFunction(GruiModel):
    """
    Every function decorated by any GruiDecorator will be replace by an instance of this class.
    Like that the GruiApp will be able to expose this function (or method if the function is define into a class)
    """
    class Argument(GruiSubModel):
        type: GruiType
        description: str
        condition: Dict[str, Any]
        optional: bool

    class Result(GruiSubModel):
        type: GruiType
        description: str

    ##############
    # Attributes #
    ##############
    id: int
    path: str
    method: str
    args: Dict[str, Argument]
    presentation: str
    description: str
    result: Optional[Result]
    depreciated: bool

    ###############
    # Constructor #
    ###############
    def __init__(self, callable_, grui_decorator):
        super().__init__()
        self.__instance = None
        self.__owner = None
        self.__properties_updated = False
        self._http_path = ""
        self._http_method = ""
        self.__args = {}
        self.__presentation = ""
        self.__description = ""
        self.__result = None
        self.default_return_code = 200  # HTTP code for OK
        self.previous_actions = []
        self.next_actions = []
        self.multiple_action = False
        if isinstance(callable_, GruiFunction):
            self.id = callable_.id
            self.__callable = callable_.__callable
            self.__grui_decorators = callable_.__grui_decorators
            if self.__properties_updated:
                raise InitialisationFunctionError
        else:
            self.id = id(callable_)
            self.__callable = callable_
            self.__grui_decorators = []
        self.__grui_decorators.append(grui_decorator)

    ###########
    # Getters #
    ###########
    @property
    def __code__(self):
        return self.__callable.__code__

    @property
    def __defaults__(self):
        return self.__callable.__defaults__

    @property
    def __kwdefaults__(self):
        return self.__callable.__kwdefaults__

    # Can not override annotation because of the grui model
    # The signature need to get the attribute of the GruiFunction
    # @property
    # def __annotations__(self):
    #     return self.__callable.__annotations__

    @property
    def __name__(self):
        if self.__owner:
            return "%s.%s.%s" % (self.__callable.__module__, self.__owner.__name__, self.__callable.__name__)
        else:
            return "%s.%s" % (self.__callable.__module__, self.__callable.__name__)

    # To trick the inspect python module we override the 5 previous properties:
    # code, defaults, kwdefaults, annotation and name.
    # Like that the method inspect.signature() return the same result as an undecorated function.

    @property
    def callable(self):
        return self.__callable

    @property
    def path(self):
        self.__update_properties()
        return "/api" + self._http_path

    @property
    def method(self):
        self.__update_properties()
        return self._http_method

    @property
    def args(self):
        self.__update_properties()
        return self.__args

    @property
    def presentation(self):
        self.__update_properties()
        return self.__presentation

    @property
    def description(self):
        self.__update_properties()
        return self.__description

    @property
    def result(self):
        self.__update_properties()
        return self.__result

    @property
    def __doc__(self):
        return self.__callable.__doc__

    @property
    def is_method(self):
        return self.__owner is not None

    @property
    def instance_class(self):
        return self.__owner

    ###########
    # Methods #
    ###########
    def __get__(self, instance, owner):
        # We save the instance of the object from the method class. In case the function is a method
        self.__instance = instance
        self.__owner = owner
        return self

    def __call__(self, *args, **kwargs):
        # Invoked on every call of any decorated method
        if self.__instance:
            return self.__callable(self.__instance, *args, **kwargs)
        else:
            return self.__callable(*args, **kwargs)

    def validate(self):
        return True

    #def to_json(self):
    #    self.__update_properties()
    #    return super().to_json()

    def __update_properties(self):
        if not self.__properties_updated:
            # Update the properties by the decorators
            for grui_decorator in self.__grui_decorators:
                grui_decorator.update_props(self)

            # Parse the docstring to get the documentation
            doc_parsed = docstring_parser.parse(self.__callable.__doc__)
            self.__presentation = doc_parsed.short_description or ""
            self.__description = doc_parsed.long_description or ""
            if doc_parsed.deprecation is None:
                self.depreciated = False

            # Save the parameter description into a temporary map.
            parameter_doc_by_name = {}
            for param_doc in doc_parsed.params:
                parameter_doc_by_name[param_doc.arg_name] = param_doc

            signature = inspect.signature(self.__callable)
            if signature.return_annotation != inspect.Signature.empty:
                self.__result = GruiFunction.Result(
                    description="" if doc_parsed.returns is None else doc_parsed.returns.description,
                    type=GruiType.get_instance(signature.return_annotation)
                )

            for param_name in signature.parameters:
                if param_name == "self":
                    continue
                annotation = self.__callable.__annotations__.get(param_name)
                parameter_doc = parameter_doc_by_name.get(param_name)
                description = ""
                optional = False
                condition = {}

                if parameter_doc is not None and annotation is not None:
                    description = parameter_doc.description
                    optional = parameter_doc.is_optional or False

                    # Parse the description to extract the condition
                    if annotation in (int, float):
                        max_value_match = re.search(MAX_ARGUMENT_PATTERN, description)
                        min_value_match = re.search(MIN_ARGUMENT_PATTERN, description)
                        if max_value_match is not None:
                            description = re.sub(MAX_ARGUMENT_PATTERN, "", description, 1).strip()
                            condition['max'] = annotation(max_value_match.group(1))
                        if min_value_match is not None:
                            description = re.sub(MIN_ARGUMENT_PATTERN, "", description, 1).strip()
                            condition['min'] = annotation(min_value_match.group(1))

                self.__args[param_name] = GruiFunction.Argument(
                    type=GruiType.get_instance(annotation),
                    description=description,
                    condition=condition,
                    optional=optional
                )

            tt = self.__callable.__name__
            if self.__callable.__name__.startswith("get") and self.__presentation == "" and \
                    self.__result is not None and self.__result.description != "":
                self.__presentation = "Retrieve " + self.__result.description[0].lower() + self.__result.description[1:]

            self.__properties_updated = True


#############
# Decorator #
#############
def _grui_decorator(default_param=()):
    def __inner(cls):
        def __wrapper(*args, **kwargs):
            if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
                # actual decorated function
                return cls(*default_param)(*args)
            else:
                # decorator arguments
                return lambda *a, **kw: cls(*args, **kwargs)(*a, **kw)
        return __wrapper
    return __inner


class _AbstractGruiDecorator(ABC):

    # This function is doing nothing but keep to avoid getting warning during Intellij inspections
    def __init__(self, *args):
        pass

    def __call__(self, action):
        return GruiFunction(action, self)

    @abstractmethod
    def update_props(self, grui_function: GruiFunction):
        pass


class _AbstractHttpRoute(_AbstractGruiDecorator):

    __REPLACE_SLASH_REGEX = re.compile("//+/")

    def __init__(self, path):
        self.path = path

    def update_props(self, grui_function: GruiFunction):
        if grui_function.is_method:
            tmp = grui_function.instance_class.url_prefix + "/" + (self.path or "")
        else:
            if self.path is None:
                tmp = snake_case(grui_function.__name__).replace(".", "-")
            else:
                tmp = self.path
        grui_function._http_path = "/" + _AbstractHttpRoute.__REPLACE_SLASH_REGEX.sub("/", tmp).strip("/")
        grui_function._http_method = upper_case(self.__class__.__name__)


@_grui_decorator((None,))
class Get(_AbstractHttpRoute):
    pass


@_grui_decorator((None,))
class Post(_AbstractHttpRoute):

    def update_props(self, grui_function: GruiFunction):
        super().update_props(grui_function)
        grui_function.default_return_code = 201  # HTTP code for Created


@_grui_decorator((None,))
class Put(_AbstractHttpRoute):
    pass


@_grui_decorator((None,))
class Delete(_AbstractHttpRoute):
    pass


@_grui_decorator((None,))
class _Grui(_AbstractHttpRoute):
    pass


@_grui_decorator(())
class NotFoundIfNone(_AbstractGruiDecorator):

    def update_props(self, grui_function: GruiFunction):
        def _(result: Any, code: int):
            if result is None:
                return None, 404  # HTTP code for Not found
            return result, code
        grui_function.next_actions.append(_)


@_grui_decorator(())
class NotFoundIfEmpty(_AbstractGruiDecorator):

    def update_props(self, grui_function: GruiFunction):
        def _(result: Any, code: int):
            if len(result) == 0:
                return None, 404  # HTTP code for Not found
            return result, code
        grui_function.next_actions.append(_)


@_grui_decorator(())
class MultipleCallAtOnce(_AbstractGruiDecorator):

    def update_props(self, grui_function: GruiFunction):
        grui_function.multiple_action = True



from os import path
import setuptools

with open(path.join(path.dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

with open(path.join(path.dirname(__file__), "requirements.txt"), "r") as fh:
    install_requires = [line.strip() for line in fh.readlines()]

setuptools.setup(
    name="grui",
    version="0.0.9",
    author="SECRET Olivier",
    author_email="pypi-package-grui@devo.live",
    description="An ready to use exposer from services oriented code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/olive007/grui",
    packages=setuptools.find_packages(),
    data_files=[('grui', ['requirements.txt'])],
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
